#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <byteswap.h>

typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned long u32;

unsigned
lzss_expand( u8 * iData, int size_p, u8 * oData, int size_u )
{
  u8            circle[0x1000];
  int           flags, distance, length, c;
  int           offset_i, offset_o, offset_c;

  offset_i = 0;			// Offset within the input buffer
  offset_o = 0;			// Offset within the output buffer
  offset_c = 0xFEE;		// Starting point in the circular buffer
  flags = 0;			// Processing flags for raw/distance+length values

  memset( circle, 0x20, 0xFEE );
  // Loop through all input bytes
  for( ;; )
  {

    // Read a new flags byte if necessary
    if( !( ( flags >>= 1 ) & 0x100 ) )
    {
      // Prevent a buffer overrun
      if( offset_i >= size_p )
	break;

      // I learned this trick from the reference implementation.
      // Flags will be > 255 if we have processed fewer than 8 bits.
      flags = 0xFF00 | ( int ) iData[offset_i++];
    }

    // Decode a raw byte
    if( flags & 1 )
    {

      // Prevent a buffer overrun
      if( offset_i >= size_p || offset_o >= size_u )
	break;

      // Copy the next byte from input to output
      oData[offset_o] = iData[offset_i++];
      circle[offset_c++] = oData[offset_o++];
      offset_c &= 0xFFF;	// Clamp to buffer range
    }

    // Decode a "distance"/length pair
    else
    {
      // Prevent a buffer overrun
      if( offset_i + 1 >= size_p )
	break;

      // Process "distance" and length values
      // Distance is actually the offset in the circular buffer
      // Bits 4-7 of length are actually bits 8-11 of distance
      distance = iData[offset_i++];
      length = iData[offset_i++];
      distance |= ( ( length & 0xF0 ) << 4 );
      length = ( length & 0x0F ) + 3;

      // Prevent a buffer overflow
      if( offset_o + length > size_u )
	break;

      // Copy bytes from the circular buffer to the output
      for( c = 0; c < length; c++ )
      {
	oData[offset_o] = circle[( distance + c ) & 0xFFF];
	circle[offset_c++] = oData[offset_o++];
	offset_c &= 0xFFF;	// Clamp to buffer range
      }
    }
  }				// for loop
  printf( "Used 0x%x input bytes; expanded to 0x%x\n", offset_i, offset_o );
  return offset_o;
}

struct
{
  u32 v1;
  u32 v2;
  u32 ilen;
  u32 zero;
} lzss_hdr;

// Program entry point
int
main( int argc, char *argv[] )
{
  int           i, len, xlen, addr;
  u8           *ibuf, *obuf;
  FILE         *fi, *fo;

  if( ( fi = fopen( "sagem.bin", "rb" ) ) == NULL )
    exit( -1 );

  fread( &lzss_hdr, 1, 0x10, fi );
  ibuf = malloc( lzss_hdr.ilen );
  fread( ibuf, 1, lzss_hdr.ilen, fi );
  fclose( fi );
  obuf = malloc( 0xF0000 );
  xlen = lzss_expand( ibuf, lzss_hdr.ilen, obuf, 0xF0000 );
  if( ( fo = fopen( "expanded.bin", "wb" ) ) != NULL )
  {
    fwrite( obuf, 1, xlen, fo );
    fclose( fo );
  }
  free( obuf );
  free( ibuf );
  return 0;
}
